/*

pdSFV 1.2 mod - February 2014

Original 1.2 by Hoopy, November 2000
  Homepage: http://rescene.wikidot.com/pdsfv
  Original homepage: http://pdsfv.isonews.com

Modifications by Joe Forster/STA <sta@c64.org>:
- 2014-02-22:
  - modified scanning of SFV lines to handle file names with spaces inside
  - removed opening file with names in different case in DOS/Windows
  - added "-q" option for Unix-style quiet mode
  - changed messages to those in the Windows/Delphi command line utilities
  - replaced end-of-lines in printed strings (in Windows, LF becomes CR+LF)
  - changed recognition of archive file extensions to more precise
  - simplified check of whether file exists and added skipping directories
  - restructured code to collect common functionality into subroutines
  - replaced static CRC32 table with a generated (once) one
  - replaced file read buffer allocation from implicit stack to explicit heap
  - added some #ifdef's for compiling with gcc/MinGW under Windows
  - got rid of compilation warnings
  - removed "Oh mama!" messages :-)
  - beautified the source with Uncrustify for consistent style

Test release: http://sta.c64.org/shared/pdsfv.zip
(to be removed upon a proper release, see: http://sta.c64.org/updates.html )

If you have GNU C installed, compile this source using

        gcc -o pdsfv pdsfv.c

or for an optimized executable compiled with gcc/MinGW under Windows

        mingw32-make

Make sure to have read the documentation that came with this source before
doing anything.

*/

#define HSFV_VERSION "pdSFV 1.2 mod"

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#ifndef _WIN32
#include <sys/mman.h>
#endif
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#if defined ( _WIN32 ) || defined ( _DOS )
#define FS_CASEINSENSITIVE 1
#define SFV_OS "DOS"
#define EOL "\n"
#else
#define SFV_OS "UNIX"
#define EOL "\r\n"
#endif

long UpdateCRC ( unsigned long CRC, const char *buffer, long count )
{
  /*
     Note: if you want to know how CRC32-checking works, I
     recommend grabbing any old (mid 90's) Z-Modem source.
     There is not much you can change in this function, so
     if you need a CRC32-check yourself, feel free to rip.
  */

  // fill the CRC32 table upon first run
  static int crc_table_computed = 0;
  static unsigned long CRCTABLE[ 256 ];
  if ( !crc_table_computed )
  {
    unsigned long c;
    int n, k;
    for ( n = 0; n < 256; n++ )
    {
      c = (unsigned long)n;
      for ( k = 0; k < 8; k++ )
      {
        if ( c & 1 )
          c = 0xedb88320L ^ ( c >> 1 );
        else
          c = c >> 1;
      }
      CRCTABLE[ n ] = c;
    }
    crc_table_computed = 1;
  }

  if ( buffer && count )
  {
    do
      CRC = ( ( CRC >> 8 ) & 0xFFFFFF ) ^ CRCTABLE[ (unsigned char)( ( CRC & 0xff ) ^ *buffer++ ) ];
    while ( --count );
  }
  return CRC;
}

long GetFileCRC ( char *filename )
{
  unsigned long crc = 0xffffffff;
  FILE *f;
  long totalread = 0;
  long localread;
  /*
     Note: different buffer sizes may result in noticable
     different performance depending on system, so feel
     free to modify.
  */
#define BUFFERSIZE 65536 * 16
  char *buffer = (char *)malloc ( BUFFERSIZE );

  if ( buffer != NULL && ( f = fopen ( filename, "rb" ) ) != NULL )
  {
    do
      if ( ( localread = fread ( buffer, 1, BUFFERSIZE, f ) ) )
      {
        crc = UpdateCRC ( crc, buffer, localread );
        totalread = totalread + localread;
      }
    while ( localread > 0 );
    fclose ( f );

    crc = crc ^ 0xffffffff;
  }
  else /* error opening file */
    crc = 0;

  free ( buffer );
  return crc;
}

int FileExists ( char *name )
{
  // return 1 for existing file, -1 for directory, 0 for missing file
  struct stat fileinfo;
  return ( stat ( name, &fileinfo ) != 0 ) ? 0 : S_ISDIR ( fileinfo.st_mode ) ? -1 : 1;
}

/* split SFV line into filename and CRC, handle filenames with spaces inside */
void ParseSFVLine ( char *sfvline, char *cfname, int *listedcrc )
{
  char *str;

  // search for the beginning of the CRC backwards
  for ( str = &sfvline[ strlen ( sfvline ) ]; str > sfvline && *( str - 1 ) != ' '; str-- )
    ;
  if ( str > sfvline )
  {
    sscanf ( str, "%X", listedcrc );

    // skip a single space before the CRC
    str--;
    if ( str > sfvline )
    {
      strncpy ( cfname, sfvline, str - sfvline );
      cfname[ str - sfvline ] = 0;
    }
  }
}

int main ( int numparams, char *params[] )
{
  int compatmode = 0;   // Win-SFV compatibility mode
  int haslogo = 0;      // add banner to created SFVs
  int quiet = 0;        // don't display unnecessary messages
  char logoname[ 250 ]; // filename of bannerfile
  char sfvname[ 250 ];  // name of .sfv file to use
  char sfvline[ 250 ];  // current line in SFV file
  int paramcnt = 1;     // command line parameter offset
  int doallfiles = 0;   // include even crappy files to SFV
  int mode = 0;         // mode of operation
  int addfinfo = 0;     // add files' dates & times to SFV

  FILE *sfvfile;
  FILE *logofile;
  time_t curtime;
  long mycrc;
  int cnt, len;
  int dothisone;
  char cfname[ 250 ];
  char crap[ 250 ];
  int numfiles = 0;
  int badfiles = 0;
  int msgfiles = 0;
  int openerror = 0;
  int listedcrc = 0;
  int fileexist;
  char mytime[ 80 ];
  int paramsave;
  struct stat fileinfo;
  struct tm zeit;

  // output program name & version number
  // output is delayed in case quiet mode is enabled
  void printtitle ( void )
  {
    static int titleprinted = 0;
    if ( !quiet && !titleprinted )
    {
      printf ( "%s\n\n", HSFV_VERSION );
      titleprinted = 1;
    }
  }

  int opensfv ( char *sfvname )
  {
    if ( !quiet )
    {
      printf ( "%s SFV: %s ...\n", ( mode == 1 ) ? "Creating" : "Processing", sfvname );
      fflush ( stdout );
    }

    sfvfile = fopen ( sfvname, ( mode == 1 ) ? "wt" : "rt" );
    if ( sfvfile == NULL )
    {
      if ( quiet )
        printf ( "%s: can't open\n", sfvname );
      else
        printf ( "can't open %s\n", sfvname );
      return 1;
    }
    return 0;
  }

  int readsfv ( char *sfvline )
  {
    while ( fgets ( sfvline, 250, sfvfile ) != NULL )
      // Skip comments, blank lines, other crap
      if ( sfvline[ 0 ] != ';' && sfvline[ 0 ] != '#' && sfvline[ 0 ] != ' ' && strlen ( sfvline ) > 8 )
        return 1;


    return 0;
  }

  // instead of reading the directory contents we're just
  // checking whether the file exists by looking for a
  // file with the filename in all lower case, all upper case
  // or first letter upper case, rest lower case. replace
  // this with a readdir(...) if you think it's needed.
  int checkfile ( char *cfname )
  {
    if ( !quiet )
    {
      printf ( "%s file %s ... ", ( mode == 1 ) ? "Processing" : "Checking", cfname );
      fflush ( stdout );
    }
    if ( FileExists ( cfname ) > 0 )
      return 1;

#ifndef FS_CASEINSENSITIVE
    for ( cnt = 0; cnt < (int)strlen ( cfname ); cnt++ )
      cfname[ cnt ] = toupper ( cfname[ cnt ] );
    if ( FileExists ( cfname ) > 0 )
      return 1;

    for ( cnt = 1; cnt < (int)strlen ( cfname ); cnt++ )
      cfname[ cnt ] = tolower ( cfname[ cnt ] );
    if ( FileExists ( cfname ) > 0 )
      return 1;

    cfname[ 0 ] = tolower ( cfname[ 0 ] );
    if ( FileExists ( cfname ) > 0 )
      return 1;

#endif
    return 0;
  }

  void countfile ( void )
  {
    if ( fileexist >= 0 ) // skip directories
    {
      if ( fileexist == 0 )
      {
        if ( quiet )
          printf ( "%s: %s\n", cfname, ( errno > 0 ) ? strerror ( errno ) : ( errno == -1 ) ? "not listed" : "" );
        else
          printf ( "MISSING!\n" );
        msgfiles++;
      }
      else
      {
        if ( mycrc == listedcrc || mode == 1 || mode == 4 )
        {
          if ( !quiet )
            printf ( "%s\n", ( mode == 1 ) ? "OK." : "GOOD!" );
        }
        else
        {
          if ( quiet )
            printf ( "%s: crc mismatch\n", cfname );
          else
            printf ( "BAD!\n" );
          badfiles++;
        }
      }
      fflush ( stdout );
      numfiles++;
    }
  }

  int printtotal ( void )
  {
    if ( badfiles > 0 )
      if ( quiet )
        printf ( "%d crc mismatches\n", badfiles );
      else
        printf ( "There were bad files!\n" );
    else
    if ( msgfiles > 0 )
      if ( quiet )
        printf ( "%s%d missing files\n", ( badfiles > 0 ) ? ", " : "", msgfiles );
      else
        printf ( "There were missing files!\n" );
    else
    if ( !quiet )
    {
      if ( numfiles > 0 )
        printf ( "%s\n", ( mode == 1 ) ? "Finished." : "All tested files were good!" );
      else
        printf ( "Nothing to do.\n" );
    }
    fflush ( stdout );

    if ( msgfiles )
      return 2;

    if ( badfiles )
      return 1;

    return 0;
  }

  // no command line paramaters given? display help!
  if ( numparams < 2 )
  {
    printtitle();
    printf ( "Usage: %s [mode | option] <sfv-filename> <filename(s)>\n\n", params[ 0 ] );
    printf ( "Possible modes: -c : create SFV file\n" );
    printf ( "                -t : test single file(s) (multiple filenames possible)\n" );
    printf ( "                -T : test whole SFV file\n" );
    printf ( "                -m : count missing files\n\n" );
    printf ( "Options:        -l <filename> : use <filename> as banner file (create mode)\n" );
    printf ( "                -w : Win-SFV emulation mode                   (create mode)\n" );
    printf ( "                -a : add all files to .sfv (even .nfo etc.)   (create mode)\n" );
    printf ( "                -d : add files' dates & times to SFV file     (create mode)\n" );
    printf ( "                -q : quiet mode\n\n" );
    printf ( "Examples:\n" );
    printf ( "(1) Create test.sfv and add test.* to it       : %s -c test.sfv test.*\n", params[ 0 ] );
    printf ( "(2) Check file test.r02 against CRC in test.sfv: %s -t test.sfv test.r02\n", params[ 0 ] );
    printf ( "(3) Check all files listed in test.sfv         : %s -T test.sfv\n", params[ 0 ] );
    printf ( "(4) Count number of files missing of test.sfv  : %s -m test.sfv\n", params[ 0 ] );
    printf ( "(5) Create myrls.sfv with myrls.* in it, use Win-SFV compatibility mode and file\n" );
    printf ( "    /tmp/mylogo.nfo as banner: %s -c myrls.sfv -w -l /tmp/mylogo.nfo myrls.*\n", params[ 0 ] );
    printf ( "\n" );
    printf ( "Read documentation files for more help and check for updates regularily!\n\n" );
    return 0;
  }

  // parse command line
  while ( paramcnt < numparams )
  {
    if ( strncmp ( params[ paramcnt ], "-", 1 ) == 0 )
      if ( ( paramcnt + 1 ) >= numparams )
      {
        printtitle();
        printf ( "Option %s needs an argument!\n", params[ paramcnt ] );
        return 1;
      }

    if ( strstr ( "-create", params[ paramcnt ] ) != NULL ) // prepare for create mode
    {
      mode = 1;
      sprintf ( sfvname, "%s", params[ paramcnt + 1 ] );
      paramcnt++;
    }
    if ( strstr ( "-test", params[ paramcnt ] ) != NULL ) // prepare single file test mode
    {
      mode = 2;
      sprintf ( sfvname, "%s", params[ paramcnt + 1 ] );
      paramcnt++;
    }
    if ( strstr ( "-TESTALL", params[ paramcnt ] ) != NULL ) // prepare sfv file test mode
    {
      mode = 3;
      sprintf ( sfvname, "%s", params[ paramcnt + 1 ] );
    }
    if ( strstr ( "-missing", params[ paramcnt ] ) != NULL ) // prepare missing file test mode
    {
      mode = 4;
      sprintf ( sfvname, "%s", params[ paramcnt + 1 ] );
    }
    if ( strstr ( "-logo", params[ paramcnt ] ) != NULL ) // set banner file name
    {
      haslogo = 1;
      sprintf ( logoname, "%s", params[ paramcnt + 1 ] );
      paramcnt++;
    }
    if ( strstr ( "-quiet", params[ paramcnt ] ) != NULL ) // quiet mode
      quiet = 1;
    if ( strstr ( "-addallfiles", params[ paramcnt ] ) != NULL ) // set lame allfiles-add-mode
      doallfiles = 1;
    if ( strstr ( "-winsfv", params[ paramcnt ] ) != NULL ) // set win-sfv compatibility mode
      compatmode = 1;
    if ( strstr ( "-dateandsizeinfo", params[ paramcnt ] ) != NULL ) // set add files' dates & times to SFV mode
      addfinfo = 1;

    paramcnt++;
    if ( paramcnt >= numparams )
      if ( mode < 3 )
      {
        printtitle();
        printf ( "No files to process! Use command without parameters for help.\n" );
        return 1;
      }

    if ( strncmp ( params[ paramcnt ], "-", 1 ) != 0 )
      break;
  }

  printtitle();
  openerror = opensfv ( sfvname );
  if ( openerror != 0 )
    return openerror;

  if ( mode == 1 )
  {
    if ( !quiet )
    {
      if ( haslogo )
        printf ( "Adding banner file        : %s\n", logoname );
      if ( compatmode )
        printf ( "Using compatibility mode  : Win-SFV\n" );
      if ( doallfiles )
        printf ( "Add-all-files mode active\n" );
      if ( addfinfo )
        printf ( "Adding files' times & dates to SFV file\n" );
      fflush ( stdout );
    }

    time ( &curtime );
    sprintf ( mytime, "%s", ctime ( &curtime ) );
    for ( cnt = 0; cnt < (int)strlen ( mytime ); cnt++ )
      if ( mytime[ cnt ] == '\n' )
        mytime[ cnt ] = '\0';

    zeit = *localtime ( &curtime );
    strftime ( mytime, 50, "%Y-%m-%d at %H:%M.%S", &zeit );

    sprintf ( sfvline, "; Generated by " HSFV_VERSION " (" SFV_OS ") on %s", mytime );
    if ( compatmode )
      sprintf ( sfvline, "; Generated by WIN-SFV32 v1 (" HSFV_VERSION ") on %s", mytime );

    fprintf ( sfvfile, "%s" EOL ";" EOL, sfvline );

    if ( haslogo )
    {
      logofile = fopen ( logoname, "rt" );
      if ( logofile == NULL )
        printf ( "couldn't find logo file %s ... omitting!\n", logoname );
      else
      {
        while ( fgets ( sfvline, 250, logofile ) != NULL )
        {
          fprintf ( sfvfile, ";!SFV_COMMENT_START" EOL ";" EOL );
          for ( cnt = 0; cnt < (int)strlen ( sfvline ); cnt++ )
            if ( sfvline[ cnt ] == '\n' )
              sfvline[ cnt ] = '\0';
          fprintf ( sfvfile, "; %s" EOL, sfvline );
        }
        fclose ( logofile );
        fprintf ( sfvfile, ";!SFV_COMMENT_END" EOL ";" EOL );
      }
    } // add logo

    if ( addfinfo )
    {
      paramsave = paramcnt;
      while ( paramsave < numparams )
      {
        stat ( params[ paramsave ], &fileinfo );

        zeit = *localtime ( &fileinfo.st_mtime );
        strftime ( mytime, 50, "%H:%M.%S %Y-%m-%d", &zeit );

        fprintf ( sfvfile, "; %12ld  %s %s" EOL, fileinfo.st_size, mytime, params[ paramsave ] );

        paramsave++;
      }
      fprintf ( sfvfile, ";" EOL );
    } // add files' times & dates to SFV file

    while ( paramcnt < numparams )
    {
      listedcrc = mycrc = 0;
      sprintf ( cfname, "%s", params[ paramcnt ] );

      // check if we should process this file: extensions "ace", "arj", "rar", "zip" or "[0amrz]##" (where # is a decimal digit)
      dothisone = 0;
      if ( doallfiles )
        dothisone = 1;
      else
      if ( strlen ( cfname ) > 4 )
      {
        char *cfext = &cfname[ strlen ( cfname ) - 4 ];
        if ( *cfext == '.' )
        {
          cfext++;
          if ( stricmp ( cfext, "ace" ) == 0
               || stricmp ( cfext, "arj" ) == 0
               || stricmp ( cfext, "rar" ) == 0
               || stricmp ( cfext, "zip" ) == 0
               || ( ( *cfext == '0' || tolower ( *cfext ) == 'a' || tolower ( *cfext ) == 'm' || tolower ( *cfext ) == 'r' || tolower ( *cfext ) == 'z' )
                    && ( cfext[ 1 ] >= '0' && cfext[ 1 ] <= '9' )
                    && ( cfext[ 2 ] >= '0' && cfext[ 2 ] <= '9' ) ) )
            dothisone = 1;
        }
      }
      if ( dothisone )
      {
        if ( ( fileexist = checkfile ( cfname ) ) >= 0 )
        {
          mycrc = GetFileCRC ( cfname );
          if ( mycrc != 0 )
          {
            sprintf ( sfvline, "%s %08lX", cfname, mycrc );
            fprintf ( sfvfile, "%s" EOL, sfvline );
          }
        }
        countfile();
      }
      paramcnt++;
    }
  } // create SFV file mode

  if ( mode == 2 )
    while ( paramcnt < numparams )
    {
      listedcrc = mycrc = 0;
      sprintf ( cfname, "%s", params[ paramcnt ] );
      if ( ( fileexist = checkfile ( cfname ) ) >= 0 )
      {
        mycrc = GetFileCRC ( cfname );

        len = strlen ( cfname );
        rewind ( sfvfile );
        dothisone = 0;

        while ( readsfv ( sfvline ) )
          if ( strnicmp ( cfname, sfvline, len ) == 0 )
          {
            dothisone = 1;
            break;
          }

        if ( dothisone )
          ParseSFVLine ( sfvline, crap, &listedcrc );
        else
        {
          fileexist = 0;
          errno = -1;  // signal file not being listed in sfv
        }
      }
      countfile();
      paramcnt++;
    }
  // test single file(s) mode

  if ( mode == 3 )
    while ( readsfv ( sfvline ) )
    {
      listedcrc = mycrc = 0;
      ParseSFVLine ( sfvline, cfname, &listedcrc );
      if ( ( fileexist = checkfile ( cfname ) ) > 0 )
        mycrc = GetFileCRC ( cfname );
      countfile();
    }
  /* test whole SFV file mode */

  if ( mode == 4 )
    while ( readsfv ( sfvline ) )
    {
      listedcrc = mycrc = 0;
      ParseSFVLine ( sfvline, cfname, &listedcrc );
      fileexist = checkfile ( cfname );
      countfile();
    }
  /* count files mode */

  fclose ( sfvfile );
  return printtotal();
}

/* EOF */
